/*	Copyright (C) 2003-2015 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#define	FE_TERM_DEBUG		FALSE
#define	FE_TERM_TIME_DEBUG	FALSE
#define	FE_TERM_COPY_DEBUG	FALSE

#include "terminal.pmh"

namespace fe
{

/***
	incoming commands:
		<result> = set "callback" <pointer>
		<result> = set <name> <variable> <type> <value>

		<result> = manage <dl>
		<result> = unmanage <dl>
		<result> = create <implementation> [name|surf]
		<result> = signal <name> [surfOut] [surfIn0] ...
		<result> = release <name>

		<result> = load <surf> <filename> [options]
		<result> = save <surf> <filename> [options]
		<result> = count <surf> point|primitive
		<result> = access <surf> point|primitive <attribute>

	callback requests:
		<value> = get <name> <variable>

	issues:
		need an abstracted get() to be used by the HandlerI instance
*/

Terminal::Terminal(void):
	m_frame(0.0)
{
}

Terminal::~Terminal(void)
{
#if FALSE
	for(std::map<String,ComponentInfo>::const_iterator it=m_surfaceMap.begin();
			it!=m_surfaceMap.end();it++)
	{
		const ComponentInfo& rComponentInfo=it->second;
		feLog("  %s %d %d\n",it->first.c_str(),
				rComponentInfo.m_spComponent.isValid(),
				rComponentInfo.m_serial);
	}
#endif
}

void Terminal::initialize(void)
{
	feRegisterSegvHandler();

	m_spScope=registry()->master()->catalog()->catalogComponent(
			"Scope","SimScope");
	FEASSERT(m_spScope.isValid());
}

String Terminal::execute(String a_commandline)
{
#if FE_TERM_DEBUG
	feLog("Terminal::execute \"%s\"\n",a_commandline.c_str());
#endif

	const String command=a_commandline.parse("\"'");

	const BWORD isGet=(command=="get");
	if(isGet || command=="set")
	{
		const String target=a_commandline.parse("\"'");
		if(target=="callback")
		{
			String pointer=a_commandline.parse("\"'");

			sscanf(pointer.c_str(),"%p",&m_fpCallback);
			return "changed callback";
		}

		if(target=="frame")
		{
			String frame=a_commandline.parse("\"'");

			sscanf(frame.c_str(),"%f",&m_frame);
			return "changed frame";
		}

		ComponentInfo& rComponentInfo=m_surfaceMap[target];
		sp<SurfaceAccessorI> spSurfaceAccessorI=rComponentInfo.m_spComponent;
		if(spSurfaceAccessorI.isValid())
		{
			String indexString=a_commandline.parse();

			I32 index=0;
			sscanf(indexString.c_str(),"%d",&index);

			const String valueString=a_commandline.parse();

			String& rTypeName=rComponentInfo.m_typeName;
			if(rTypeName.empty())
			{
				rTypeName=spSurfaceAccessorI->type();
			}
//			feLog("target %s type \"%s\"\n",target.c_str(),rTypeName.c_str());

			if(rTypeName=="string")
			{
				if(isGet)
				{
					return spSurfaceAccessorI->string(index);
				}
				spSurfaceAccessorI->set(index,valueString);
			}
			if(rTypeName=="integer")
			{
				if(isGet)
				{
					return print(spSurfaceAccessorI->integer(index));
				}
				const I32 value=atol(valueString.c_str());
				spSurfaceAccessorI->set(index,value);
			}
			if(rTypeName=="real")
			{
				if(isGet)
				{
					return print(spSurfaceAccessorI->real(index));
				}
				const Real value=atof(valueString.c_str());
				spSurfaceAccessorI->set(index,value);
			}
			if(rTypeName=="vector3")
			{
				if(isGet)
				{
					return print(spSurfaceAccessorI->spatialVector(index));
				}
				Real x=Real(0);
				Real y=Real(0);
				Real z=Real(0);
				sscanf(valueString.c_str(),"%f%f%f",&x,&y,&z);

				SpatialVector value(x,y,z);

				spSurfaceAccessorI->set(index,value);
			}
			if(rTypeName=="vertices")
			{
				if(isGet)
				{
					String result;

					const I32 subCount=spSurfaceAccessorI->subCount(index);
					for(I32 subIndex=0;subIndex<subCount;subIndex++)
					{
						String number;
						number.sPrintf("%d",
								spSurfaceAccessorI->integer(index,subIndex));

						result+=(subIndex? " ": "") + number;
					}

					return result;
				}

				//* TODO set vertices
			}

			return "";
		}

		if(m_nodeMap.find(target)==m_nodeMap.end())
		{
			return "error: unknown target";
		}

		sp<Catalog> spCatalog=m_nodeMap[target].catalog();

		const String variable=a_commandline.parse("\"'");
		if(variable.empty())
		{
			return "error: missing variable name";
		}

//		feLog("variable \"%s\"\n",variable.c_str());

		if(isGet)
		{
			if(!spCatalog->cataloged(variable))
			{
				return "error: missing variable";
			}

			const Instance& rInstance=spCatalog->catalogInstance(variable);

			const sp<BaseType> spBaseType=rInstance.type();
			if(spBaseType.isNull())
			{
				return "error: no type";
			}

			const TypeInfo& rTypeInfo=spBaseType->typeinfo();

			String result;

			if(rTypeInfo==typeid(String))
			{
				result=spCatalog->catalog<String>(variable);
			}
			else if(rTypeInfo==typeid(bool))
			{
				result.sPrintf("%d",spCatalog->catalog<bool>(variable));
			}
			else if(rTypeInfo==typeid(I32))
			{
				result.sPrintf("%d",spCatalog->catalog<I32>(variable));
			}
			else if(rTypeInfo==typeid(Real))
			{
				result.sPrintf("%G",spCatalog->catalog<Real>(variable));
			}
			else if(rTypeInfo==typeid(SpatialVector))
			{
				result.sPrintf("%s",
						c_print(spCatalog->catalog<SpatialVector>(variable)));
			}
			else if(rTypeInfo==typeid(sp<Component>))
			{
				result.sPrintf("%p",
						spCatalog->catalog< sp<Component> >(variable).raw());
			}
			else
			{
				result="error: unrecognized type";
			}
			return result;
		}
		else
		{
			const String type=a_commandline.parse("\"'");
			if(type.empty())
			{
				return "error: missing type";
			}
//			feLog("type \"%s\"\n",type.c_str());

			const String value=a_commandline.parse("\"'");
			if(value.empty() && type!="string")
			{
				return "error: missing value";
			}
//			feLog("value \"%s\"\n",value.c_str());

			if(type=="string")
			{
				spCatalog->catalog<String>(variable)=value;
			}
			else if(type=="boolean")
			{
				spCatalog->catalog<bool>(variable)=
						(value.integer()!=0 || value=="TRUE" ||
						value=="True" || value=="true");
			}
			else if(type=="integer")
			{
				spCatalog->catalog<I32>(variable)=value.integer();
			}
			else if(type=="real")
			{
				spCatalog->catalog<Real>(variable)=value.real();
			}
			else if(type=="vector3")
			{
				Real x=Real(0);
				Real y=Real(0);
				Real z=Real(0);
				sscanf(value.c_str(),"%f%f%f",&x,&y,&z);
				spCatalog->catalog<SpatialVector>(variable)=
						SpatialVector(x,y,z);
			}
			else
			{
				return "error: unknown type";
			}
			return "changed catalog";
		}
	}

	if(command=="manage")
	{
		const String name=a_commandline.parse("\"'");

		Result result=registry()->manage(name);
		if(successful(result))
		{
			return "managed";
		}
		return "error: unknown library";
	}

	if(command=="unmanage")
	{
		const String name=a_commandline.parse("\"'");

		Result result=registry()->unmanage(name);
		if(successful(result))
		{
			return "unmanaged";
		}
		return "error: unknown library";
	}

	if(command=="create")
	{
		const String implementation=a_commandline.parse("\"'");
		if(implementation.empty())
		{
			return "error: missing implementation";
		}

		const String name=a_commandline.parse("\"'");
		if(name.empty())
		{
			return "error: missing name";
		}

		m_nodeMap[name].setupOperator(m_spScope,implementation,name);

		return "created";
	}

	if(command=="release")
	{
		const String name=a_commandline.parse("\"'");

		if(m_nodeMap.find(name)!=m_nodeMap.end())
		{
			m_nodeMap.erase(name);
			return "released node";
		}
		else if(m_surfaceMap.find(name)!=m_surfaceMap.end())
		{
#if FE_TERM_COPY_DEBUG
			feLog("Terminal::execute release \"%s\"\n",name.c_str());
#endif

			m_surfaceMap.erase(name);
			return "released surface";
		}

		return "error: unknown target";
	}

	if(command=="signal")
	{
#if	FE_TERM_TIME_DEBUG
		const U32 us1=systemMicroseconds();
#endif

		const String instance=a_commandline.parse("\"'");

		sp<Catalog> spCatalog=m_nodeMap[instance].catalog();
		FEASSERT(spCatalog.isValid());

		BWORD recycle=FALSE;
		String copy;
		String outputName;
		String outputKey;

		BWORD copied=FALSE;

#if FE_TERM_COPY_DEBUG
		feLog("Terminal::execute \"%s\"\n",instance.c_str());
#endif

		I32 surfaceIndex= -1;
		while(TRUE)
		{
			sp<SurfaceAccessibleI> spSurfaceAccessibleI;

			const String surfaceName=a_commandline.parse("\"'");
			if(surfaceIndex<0)
			{
				outputName=surfaceName;
			}
			if(surfaceName.empty())
			{
				if(a_commandline.empty())
				{
					break;
				}
			}
			else if(m_surfaceMap.find(surfaceName)!=m_surfaceMap.end())
			{
				spSurfaceAccessibleI=
						m_surfaceMap[surfaceName].m_spComponent;
			}
			else if(surfaceIndex>=0)
			{
				return "error: unknown surface \"" + surfaceName + "\"";
			}

			if(surfaceIndex>=0 && spSurfaceAccessibleI.isNull())
			{
				return "error: invalid surface \"" + surfaceName + "\"";
			}

			//* multiple inputs (named by index)

			String lookup="output:0";
			if(surfaceIndex>=0)
			{
				lookup.sPrintf("input:%d",surfaceIndex);
			}
			const String key=spCatalog->catalog<String>(lookup);

			if(surfaceIndex<0)
			{
				if(spCatalog->cataloged(key))
				{
					copy=spCatalog->catalogOrDefault<String>(key,
							"copy","");

					recycle=spCatalog->catalogOrDefault<bool>(key,
							"recycle",false);
				}

#if FE_TERM_COPY_DEBUG
				feLog("  output \"%s\" copies \"%s\" recycle %d\n",
						key.c_str(),copy.c_str(),recycle);
#endif
			}
			else
			{
				ComponentInfo& rComponentInfo=m_surfaceMap[outputName];
				std::vector<String>& rInputKey=rComponentInfo.m_inputKey;
				std::vector<U32>& rInputSerial=rComponentInfo.m_inputSerial;

				if(I32(rInputKey.size())<surfaceIndex+1)
				{
					rInputKey.resize(surfaceIndex+1);
					rInputSerial.resize(surfaceIndex+1);
				}

				String& rLastKey=rInputKey[surfaceIndex];
				U32& rLastSerial=rInputSerial[surfaceIndex];

				const U32 serial=m_surfaceMap[surfaceName].m_serial;

				const BWORD replaced=
						(key!=rLastKey || serial!=rLastSerial);
				if(replaced && recycle && key==copy)
				{
					recycle=FALSE;

#if FE_TERM_COPY_DEBUG
					feLog("  recycle off\n");
#endif
				}

#if FE_TERM_COPY_DEBUG
				feLog("  input \"%s\"/\"%s\" %d/%d replaced %d\n",
						key.c_str(),rLastKey.c_str(),
						serial,rLastSerial,replaced);
#endif

				spCatalog->catalog<bool>(key,"replaced")=replaced;

				rLastKey=key;
				rLastSerial=serial;
			}

//			feLog("surface \"%s\" is really \"%s\"\n",
//					lookup.c_str(),key.c_str());

			spCatalog->catalog< sp<Component> >(key)=spSurfaceAccessibleI;

			//* TODO respect recycle
			if(surfaceIndex>=0 && key==copy)
			{
				ComponentInfo& rComponentInfo=m_surfaceMap[outputName];

				if(spSurfaceAccessibleI.isValid() && (!recycle ||
						rComponentInfo.m_spComponent.isNull()))
				{
#if FE_TERM_COPY_DEBUG
					feLog("  clone \"%s\" \"%s\" to \"%s\"\n",
							FE_TYPESTRING(*spSurfaceAccessibleI.raw()).c_str(),
							key.c_str(),outputKey.c_str());
#endif

					sp<SurfaceAccessibleI> spOutputAccessible(
							dynamic_cast<SurfaceAccessibleI*>(
							spSurfaceAccessibleI->clone()));
					FEASSERT(spOutputAccessible.isValid());

					spOutputAccessible->setName(outputName);

					spCatalog->catalog< sp<Component> >(outputKey)=
							spOutputAccessible;

					rComponentInfo.m_serial++;
					rComponentInfo.m_spComponent=spOutputAccessible;
					copied=TRUE;
				}
			}
			else if(surfaceIndex<0)
			{
				outputKey=key;
			}

			surfaceIndex++;
		}

		if(!copied)
		{
			sp<SurfaceAccessibleI> spOutputAccessible=
					m_surfaceMap[outputName].m_spComponent;

			if(!recycle || spOutputAccessible.isNull())
			{
#if FE_TERM_COPY_DEBUG
			feLog("  new surface recycle %d null %d\n",
					recycle,spOutputAccessible.isNull());
#endif

				sp<Registry> spRegistry=m_spScope->registry();

				//* Record based
				spOutputAccessible=
						spRegistry->create("*.SurfaceAccessibleRecord");
				if(spOutputAccessible.isValid())
				{
					sp<Scope> spScope=registry()->create("Scope");
					spOutputAccessible->bind(spScope);
					spOutputAccessible->clear();
				}

				if(spOutputAccessible.isNull())
				{
					//* NOTE not verified

					//* json based
					sp<SurfaceAccessibleI> spOutputAccessible=
							spRegistry->create("SurfaceAccessibleI.*.*.geo");

					if(spOutputAccessible.isValid())
					{
						//* initialize empty
						spOutputAccessible=dynamic_cast<SurfaceAccessibleI*>(
								spOutputAccessible->clone());
					}
				}

				if(spOutputAccessible.isNull())
				{
					return "error: failed to create output surface";
				}

				spOutputAccessible->setName(outputName);

				spCatalog->catalog< sp<Component> >(outputKey)=
						spOutputAccessible;

				m_surfaceMap[outputName].m_spComponent=
						spOutputAccessible;
			}
		}

#if	FE_TERM_TIME_DEBUG
		const U32 us2=systemMicroseconds();
#endif

		const BWORD success=m_nodeMap[instance].cook(m_frame);

#if	FE_TERM_TIME_DEBUG
		const U32 us3=systemMicroseconds();
#endif

		if(!success)
		{
#if FE_TERM_COPY_DEBUG
			feLog("Terminal::execute erase \"%s\"\n",outputName.c_str());
#endif

			m_surfaceMap.erase(outputName);
		}

/*
		// TEMP make request directly
		if(!m_fpCallback)
		{
			return "error: no callback";
		}

		String request;
		request.sPrintf("get %s args",instance.c_str());

#if FE_TERM_DEBUG
		feLog("Terminal::execute request \"%s\"\n",request.c_str());
#endif

		const std::string response=m_fpCallback(request.c_str());

#if FE_TERM_DEBUG
		feLog("Terminal::execute response \"%s\"\n",response.c_str());
#endif
*/

#if	FE_TERM_TIME_DEBUG
		const U32 us4=systemMicroseconds();
		feLog("signal cook %s %d/%d us\n",instance.c_str(),us3-us2,us4-us1);
#endif

		return success? "signaled": "error: signal failed";
	}

	if(command=="load")
	{
		const String surfaceName=a_commandline.parse("\"'");
		if(surfaceName.empty())
		{
			return "error: surface name missing";
		}

		const String filename=a_commandline.parse("\"'");
		if(filename.empty())
		{
			return "error: filename missing";
		}

		const String options=a_commandline.parse("\"'");

		const String suffix=String(strrchr(filename.c_str(),'.')).prechop(".");

		ComponentInfo& rComponentInfo=m_surfaceMap[surfaceName];
		rComponentInfo.m_serial++;

		sp<SurfaceAccessibleI> spSurfaceAccessibleI=
				rComponentInfo.m_spComponent;

		sp<Catalog> spSettings=
				registry()->master()->createCatalog("Load Settings");
		spSettings->catalog<Real>("frame")=m_frame;

		if(!options.empty())
		{
			spSettings->catalog<String>("options")=options;
		}

		sp<Registry> spRegistry=m_spScope->registry();
		spSurfaceAccessibleI=
				spRegistry->create("SurfaceAccessibleI.*.*."+suffix);
		rComponentInfo.m_spComponent=spSurfaceAccessibleI;

		if(spSurfaceAccessibleI.isNull())
		{
			return "error: create driver failed";
		}

		spSurfaceAccessibleI->setName(surfaceName);
		spSurfaceAccessibleI->bind(sp<Scope>(spRegistry->create("Scope")));

		const BWORD success=spSurfaceAccessibleI->load(filename,spSettings);

		return success? "loaded": "error: load failed";
	}

	if(command=="save")
	{
		const String surfaceName=a_commandline.parse("\"'");
		if(surfaceName.empty())
		{
			return "error: surface name missing";
		}

		if(m_surfaceMap.find(surfaceName)==m_surfaceMap.end())
		{
			return "error: unknown surface";
		}

		const String filename=a_commandline.parse("\"'");
		if(filename.empty())
		{
			return "error: filename missing";
		}

		const String options=a_commandline.parse("\"'");

		ComponentInfo& rComponentInfo=m_surfaceMap[surfaceName];

		if(rComponentInfo.m_spComponent.isNull())
		{
			return "error: null surface";
		}

		const String suffix=String(strrchr(filename.c_str(),'.')).prechop(".");

		sp<Registry> spRegistry=m_spScope->registry();
		sp<SurfaceAccessibleI> spSaveAccessible=
				spRegistry->create("SurfaceAccessibleI.*.*."+suffix);

		sp<Catalog> spSettings=
				registry()->master()->createCatalog("Load Settings");
		spSettings->catalog<Real>("frame")=m_frame;

		if(options.empty())
		{
			spSettings->catalog<String>("options")=options;
		}

		spSaveAccessible->bind(m_spScope);
		spSaveAccessible->copy(rComponentInfo.m_spComponent);

		const BWORD success=spSaveAccessible->save(filename,spSettings);

		return success? "saved": "error: save failed";
	}

	if(command=="count")
	{
		const String surfaceName=a_commandline.parse();
		if(surfaceName.empty() ||
				m_surfaceMap.find(surfaceName)==m_surfaceMap.end())
		{
			return "0";
		}

		const String elementName=a_commandline.parse();
		if(elementName.empty() ||
				!(elementName=="point" || elementName=="primitive"))
		{
			return "0";
		}

		ComponentInfo& rComponentInfo=m_surfaceMap[surfaceName];
		if(rComponentInfo.m_spComponent.isNull())
		{
			return "0";
		}

		sp<SurfaceAccessibleI> spSurfaceAccessibleI=
				rComponentInfo.m_spComponent;

		String result;
		result.sPrintf("%d",spSurfaceAccessibleI->count(
				elementName=="primitive"?
				SurfaceAccessibleI::e_primitive:
				SurfaceAccessibleI::e_point));

		return result;
	}

	if(command=="specs")
	{
		const String surfaceName=a_commandline.parse();
		if(surfaceName.empty() ||
				m_surfaceMap.find(surfaceName)==m_surfaceMap.end())
		{
			return "";
		}

		const String elementName=a_commandline.parse();
		if(elementName.empty() ||
				!(elementName=="point" || elementName=="primitive"))
		{
			return "";
		}

		ComponentInfo& rComponentInfo=m_surfaceMap[surfaceName];
		if(rComponentInfo.m_spComponent.isNull())
		{
			return "";
		}

		sp<SurfaceAccessibleI> spSurfaceAccessibleI=
				rComponentInfo.m_spComponent;

		SurfaceAccessibleI::Element element=
				(elementName=="primitive")?
				SurfaceAccessibleI::e_primitive:
				SurfaceAccessibleI::e_point;

		String result;

		Array<SurfaceAccessibleI::Spec> specs;
		spSurfaceAccessibleI->attributeSpecs(specs,element);

		const U32 specCount=specs.size();
		for(U32 specIndex=0;specIndex<specCount;specIndex++)
		{
			const SurfaceAccessibleI::Spec& spec=specs[specIndex];
			const String& specName=spec.name();
			const String& specType=spec.typeName();

			result+=(specIndex? ",": "")+specType+" "+specName;

//			feLog("spec %d/%d \"%s\" \"%s\"\n",specIndex,specCount,
//					specType.c_str(),specName.c_str());
		}

		return result;
	}

	if(command=="access")
	{
		const String surfaceName=a_commandline.parse();
		if(surfaceName.empty() ||
				m_surfaceMap.find(surfaceName)==m_surfaceMap.end())
		{
			return "";
		}

		const String elementName=a_commandline.parse();
		if(elementName.empty() ||
				!(elementName=="point" || elementName=="primitive"))
		{
			return "";
		}

		const String attribute=a_commandline.parse();
		if(attribute.empty())
		{
			return "";
		}

		ComponentInfo& rSurfaceInfo=m_surfaceMap[surfaceName];
		if(rSurfaceInfo.m_spComponent.isNull())
		{
			return "";
		}

		sp<SurfaceAccessibleI> spSurfaceAccessibleI=
				rSurfaceInfo.m_spComponent;
		if(spSurfaceAccessibleI.isNull())
		{
			return "";
		}

		const String accessorName=surfaceName+":"+elementName+":"+attribute;
		ComponentInfo& rAccessorInfo=m_surfaceMap[accessorName];
		rAccessorInfo.m_serial++;

		SurfaceAccessibleI::Element element=
				(elementName=="primitive")?
				SurfaceAccessibleI::e_primitive:
				SurfaceAccessibleI::e_point;

		Array<SurfaceAccessibleI::Spec> specs;
		spSurfaceAccessibleI->attributeSpecs(specs,element);

		const U32 specCount=specs.size();
		for(U32 specIndex=0;specIndex<specCount;specIndex++)
		{
			const SurfaceAccessibleI::Spec& spec=specs[specIndex];
			const String& specName=spec.name();
			const String& specType=spec.typeName();

			if(specName==attribute)
			{
				rAccessorInfo.m_typeName=specType;
			}
		}

		sp<SurfaceAccessorI> spSurfaceAccessorI;

		if(attribute=="<vertices>")
		{
			rAccessorInfo.m_typeName="vertices";

			spSurfaceAccessorI=spSurfaceAccessibleI->accessor(element,
					SurfaceAccessibleI::e_vertices);
		}
		else
		{
			spSurfaceAccessorI=
					spSurfaceAccessibleI->accessor(element,attribute);
		}

		rAccessorInfo.m_spComponent=spSurfaceAccessorI;

		if(spSurfaceAccessorI.isNull())
		{
			return "";
		}

		return accessorName;
	}

	return "error: unknown command";
}

} // namespace
